/*
******************************************************************************
**  CarMaker - Version 7.0.2
**  Vehicle Dynamic Simulation Toolkit
**
**  Copyright (C)   IPG Automotive GmbH
**					Bannwaldallee 60             Phone  +49.721.98520.0
**                  76185 Karlsruhe              Fax    +49.721.98520.99
**                  Germany                      WWW    http://www.ipg.de
******************************************************************************
**
**  Functions, Variables, Types
**  ---------------------------
**
** tExtendedAeroParam
** ExtendedAero
** ExtendedAero_DeclQuants ()
** ExtendedAero_New ()
** ExtendedAero_Calc ()
** ExtendedAero_Delete ()
** Aero_Register_ExtendedAero ()
**
** The register function is declared in ExtendedAero.h.
** ExtendedAero.h is included in the User.c and will be called in 
** User_Register()
**
** int Aero_Register_ExtendedAero (void);
**
******************************************************************************
** This File is an example for calculating the lift and drag forces on
** the front and rear axle, depending on the ride height of the vehicle.
** It is assumed, that the approach angle is alway zero. So there are no 
** aerodynamic side forces. 
** For each axle and each force (lift and drag) there are three coefficients:
** A static, one depending on the ride height of the axle, respectively, and one
** depending on the ride height of both axles.
** The coefficients will be looked up in 1-D or 2-D maps.
** The Vehicle "DemoCar_RideHeightAero" is prepared with those maps and other necessary
** parameters for this example. The parameters are on the sheet Misc in the
** sector additional parameters of the vehicle parameter file. Here is an abstract
** of the file:

## Additional Parameters #################################################
** The approach area:
Aero.Ax = 2.0

** the 2D Map of coefficients:
** the rows depend on the front ride height, 
** the columns depend on the rear ride height
** the number of rows and columns has to be the same as the 
** vectorlength of ride height positions. Actually the length is
** limited to 30 values (see maxlen_vrh_f below).

** 2D-map for the drag coefficient
Aero.Map2D_c_x:
	0.0	0.0	0.0	0.0
	0.0	0.0	0.0	0.0
	0.0	0.0	0.0	0.0
	
** 2D-map for the lift coefficient at the front
Aero.Map2D_c_zf:
	-0.6	-0.7	-1.2	-2.0
	-0.1	-0.25	-0.4	-0.9
	0.1	-0.1	-0.3	-0.45
	
** 2D-map for the lift coefficient at the rear
Aero.Map2D_c_zr:
	-0.65	-0.75	-1.4	-2.25
	-0.15	-0.35	-0.5	-1.15
	0.1	-0.2	-0.45	-0.6

** 1D-maps for the coefficients 
Aero.Map_c_zf = 0.0	0.0	0.0
Aero.Map_c_zr = 0.0	0.0	0.0

** static drag and lift coefficients
Aero.c_x0  = 0.15
Aero.c_zf0 = 0.1
Aero.c_zr0 = 0.1

** Point of attack in the CarMaker models (The forces on the axles have
** to be calculated to forces and torques on this alternative point of attack)
Aero.pos_CarMaker = 2.25 0.0 0.6
** Front point of attack 
Aero.pos_f = 2.25 0.0 0.6
** Rear point of attack
Aero.pos_r = 2.25 0.0 0.6
** Offset of the ride height, equivalent to the distance of the wheelcenter to the measuring
** point of the ride height
Aero.rh_offset_f = 0.22
Aero.rh_offset_r = 0.22
** Time of the low pass Filter to simulate the hysteretic variation
Aero.t_Frc = 0.05

** ride height positions for the coefficient maps
Aero.vector_rh_f_length = 3
Aero.vector_rh_r_length = 4
Aero.vector_rh_f = 0.02 	0.05 	0.11
Aero.vector_rh_r = 0.015	0.035	0.08	0.11

** 
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include "infoc.h"
#include "Global.h"
#include "InfoUtils.h"
#include "EnvUtils.h"
#include <CarMaker.h>
#include <ModelManager.h>
#include <Car/Car.h>
#include <Car/Aero.h>
#include <Car/MBSUtils.h>
#include "Log.h"
#include "LinMap.h"

#include "ExtendedAero.h"

/* Defining the length of the ride-height-position-vector */
#define maxlen_vec_rhf 30
#define maxlen_vec_rhr 30

static const char ThisModelKind[] =	"ExtendedAero";
static const int  ThisVersionId =	1;

struct tVhclPart {
    double k0,k1,k2,k3;
};	

struct tExtendedAero {
    double Ax;
    double PoA_CM[3], PoA_f[3], PoA_r[3];
    float rh, rh_f, rh_r;			/* ride height on front and rear axle */
    float v_rhf, v_rhr, v_rh;  			/* velocity of ride height */
    double rh_offset_f, rh_offset_r;		/* ride height offset: distance*/
    double Frc_x;				/* Drag forces */
    double Frc_zf, Frc_zr;  			/* Lift forces on rear axle */
    double t_Frcx, t_Frczf, t_Frczr;
    double c_x0, c_x2;
    double c_zf0, c_zf1, c_zf2;
    double c_zr0, c_zr1, c_zr2;
    double c_x_tot;
    double c_zf_tot;
    double c_zr_tot;
    struct tVhclPart Drag,LiftF,LiftR;
    tLM    *Mapc_zf, *Mapc_zr;			/* CarMaker function for linear interpolation of 1D map points */
    tLM2D  *Map2Dc_x, *Map2Dc_zf, *Map2Dc_zr; 	/* CarMaker function for linear interpolation of 2D map points */
};


double rhf_old, rhr_old, rh_old;
double Frc_x_old;
double Frc_zf_old, Frc_zr_old;

/*
** ExtendedAero_DeclQuants ()
**
*/

static void
ExtendedAero_DeclQuants (void *MP)
{
    struct tExtendedAero *mp;
    static struct tExtendedAero ExtendedAero_Dummy;

    /* if MP == NULL, use static variables
     * else use dynamically allocated and actually active variables
     */
    mp = (MP != NULL ? (struct tExtendedAero*)MP : &ExtendedAero_Dummy);
	
    DDefFloat(NULL, "Aero.RideHeightfront",	"m",	&mp->rh_f, DVA_None);
    DDefFloat(NULL, "Aero.RideHeightrear",	"m",	&mp->rh_r, DVA_None);
    DDefFloat(NULL, "Aero.v_RideHeightfront", "m/s",	&mp->v_rhf, DVA_None);
    DDefFloat(NULL, "Aero.v_RideHeightrear",	"m/s",	&mp->v_rhr, DVA_None);
    DDefDouble(NULL, "Aero.Frc_drag",		"N",	&mp->Frc_x, DVA_None);
    DDefDouble(NULL, "Aero.Frc_lift_f",		"N",	&mp->Frc_zf, DVA_None);
    DDefDouble(NULL, "Aero.Frc_lift_r",		"N",	&mp->Frc_zr, DVA_None);
    DDefDouble(NULL, "Aero.t_Frc_lift_f",	"s",	&mp->t_Frczf, DVA_None);
    DDefDouble(NULL, "Aero.t_Frc_lift_r",	"s",	&mp->t_Frczr, DVA_None);
    DDefDouble(NULL, "Aero.t_Frc_drag",		"s",	&mp->t_Frcx, DVA_None);
	
}



/*
** ExtendedAero_New ()
**
** Return:
** NULL    Error
** !=NULL  ok, handle to this model and it's paramters    
**
*/

static void *
ExtendedAero_New (tInfos *Inf, const char *KindKey)
{
    struct tExtendedAero *mp = NULL;
    double t_Frc;
    char *key;
    double dvec[maxlen_vec_rhf*maxlen_vec_rhr];
    float  fvec[maxlen_vec_rhf*maxlen_vec_rhr];
    float vec_rhf[maxlen_vec_rhf];
    float vec_rhr[maxlen_vec_rhr];
    int k, len_vec_rhf, len_vec_rhr, n, m;
    
    unsigned	nError =   GetInfoErrorCount() + Log_nError;

    rh_old  = 0.0;
    rhf_old = 0.0;
    rhr_old = 0.0;
    Frc_x_old = 0.0;
    Frc_zf_old = 0.0;
    Frc_zr_old = 0.0;

    mp = (struct tExtendedAero*)calloc(1,sizeof(*mp));

    mp->t_Frcx  = 0.001;
    mp->t_Frczf = 0.001;
    mp->t_Frczr = 0.001;
    
    /* Read the aerodynamic parameters*/
    /***********************************/
    mp->Ax = iGetDbl(Inf, "Aero.Ax");
    iGetTable (Inf, "Aero.pos_CarMaker", mp->PoA_CM, 3, 1, &n);
    iGetTable (Inf, "Aero.pos_f", mp->PoA_f, 3, 1, &n);
    iGetTable (Inf, "Aero.pos_r", mp->PoA_r, 3, 1, &n);
    mp->rh_offset_f = iGetDbl(Inf, "Aero.rh_offset_f");
    mp->rh_offset_r = iGetDbl(Inf, "Aero.rh_offset_r");


    t_Frc  = iGetDblOpt(Inf, "Aero.t_Frc_k0", 0.0);
    if (t_Frc == 0.0) {
	mp->Drag.k0  = iGetDbl(Inf, "Aero.t_Frc_drag_k0");
	mp->LiftF.k0 = iGetDbl(Inf, "Aero.t_Frc_liftF_k0");
	mp->LiftR.k0 = iGetDbl(Inf, "Aero.t_Frc_liftR_k0");
    }else{
	mp->Drag.k0  = t_Frc;
	mp->LiftF.k0 = t_Frc;
	mp->LiftR.k0 = t_Frc;
    }
    mp->Drag.k1  = iGetDblOpt(Inf, "Aero.t_Frc_drag_k1",0.0);
    mp->LiftF.k1 = iGetDblOpt(Inf, "Aero.t_Frc_liftF_k1",0.0);
    mp->LiftR.k1 = iGetDblOpt(Inf, "Aero.t_Frc_liftR_k1",0.0);
    
    mp->Drag.k2  = iGetDblOpt(Inf, "Aero.t_Frc_drag_k2",0.0);
    mp->LiftF.k2 = iGetDblOpt(Inf, "Aero.t_Frc_liftF_k2",0.0);
    mp->LiftR.k2 = iGetDblOpt(Inf, "Aero.t_Frc_liftR_k2",0.0);
    
    mp->Drag.k3  = iGetDblOpt(Inf, "Aero.t_Frc_drag_k3",0.0);
    mp->LiftF.k3 = iGetDblOpt(Inf, "Aero.t_Frc_liftF_k3",0.0);
    mp->LiftR.k3 = iGetDblOpt(Inf, "Aero.t_Frc_liftR_k3",0.0);


    mp->c_x0 = iGetDbl(Inf, "Aero.c_x0");
    mp->c_zf0 = iGetDbl(Inf, "Aero.c_zf0");
    mp->c_zr0 = iGetDbl(Inf, "Aero.c_zr0");
    
    /***********************************/
    /* Read vector ride height front */
    key = "Aero.vector_rh_f";
    k = iGetTable(Inf, key, dvec, maxlen_vec_rhf, 1, &len_vec_rhf);
    if (k==0) {
	for (k=0; k < len_vec_rhf; k++) {
	    vec_rhf[k] = (float)dvec[k];
	}
    } else {
	nError++;
    }

    /* Read vector ride height rear */
    key = "Aero.vector_rh_r";
    k = iGetTable(Inf, key, dvec, maxlen_vec_rhr, 1, &len_vec_rhr);
    if (k==0) {
	for (k=0; k < len_vec_rhr; k++) {
	    vec_rhr[k] = (float)dvec[k];
	}
    } else {
	nError++;
    }

    /***********************************/
    /* Read 1D-maps with coefficients: */

    key = "Aero.Map_c_zf";
    k = iGetTable(Inf, key, dvec, maxlen_vec_rhf, 1, &n);

    if (k==0) {
	for (k=0; k < n; k++) {
	    fvec[k] = (float)dvec[k];
	}
	mp->Mapc_zf = LMInit(vec_rhf, fvec, len_vec_rhf, ArrayNotAequi);
	if ( mp->Mapc_zf == NULL) {
	    LogErrF (EC_Init, "No Mapping for '%s'", key);
	}
    } else {
	nError++;
    }
    
    key = "Aero.Map_c_zr";
    k = iGetTable(Inf, key, dvec, maxlen_vec_rhr, 1, &n);

    if (k==0) {
	for (k=0; k < n; k++) {
	    fvec[k] = (float)dvec[k];
	}
	mp->Mapc_zr = LMInit(vec_rhr, fvec, len_vec_rhr, ArrayNotAequi);
	if ( mp->Mapc_zr == NULL) {
	    LogErrF (EC_Init, "No Mapping for '%s'", key);
	}
    } else {
	nError++;
    }
    
    
    /***********************************/
    /* Read 2D-maps with coefficients: */
    
    key = "Aero.Map2D_c_x";
    k = iGetTable(Inf, key, dvec, maxlen_vec_rhf * maxlen_vec_rhr, len_vec_rhr, &n);

    if (k==0) {
	m = len_vec_rhr*n;
	for (k=0; k < m; k++) {
	    fvec[k] = (float)dvec[k];
	}
	mp->Map2Dc_x = LM2DInit(vec_rhf, len_vec_rhf, vec_rhr, len_vec_rhr, fvec, ArrayNotAequi,
				   ArrayNotAequi);
	if ( mp->Map2Dc_x == NULL) {
	    LogErrF (EC_Init, "No Mapping for '%s'", key);
	}
    } else {
	nError++;
    }

    key = "Aero.Map2D_c_zf";
    k = iGetTable(Inf, key, dvec, maxlen_vec_rhf * maxlen_vec_rhr, len_vec_rhr, &n);

    if (k==0) {
	m = len_vec_rhr*n;
	for (k=0; k < m; k++) {
	    fvec[k] = (float)dvec[k];
	}
	mp->Map2Dc_zf = LM2DInit(vec_rhf, len_vec_rhf, vec_rhr, len_vec_rhr, fvec, ArrayNotAequi,
				 ArrayNotAequi);
	if ( mp->Map2Dc_zf == NULL) {
	    LogErrF (EC_Init, "No Mapping for '%s'", key);
	}
    } else {
	nError++;
    }
    
    key = "Aero.Map2D_c_zr";
    k = iGetTable(Inf, key, dvec, maxlen_vec_rhf * maxlen_vec_rhr, len_vec_rhr, &n);

    if (k==0) {
	m = len_vec_rhr*n;
	for (k=0; k < m; k++) {
	    fvec[k] = (float)dvec[k];
	}
	mp->Map2Dc_zr = LM2DInit(vec_rhf, len_vec_rhf, vec_rhr, len_vec_rhr, fvec, ArrayNotAequi,
				 ArrayNotAequi);
	if ( mp->Map2Dc_zr == NULL) {
	    LogErrF (EC_Init, "No Mapping for '%s'", key);
	}
    } else {
	nError++;
    }
    

    if (nError != GetInfoErrorCount() + Log_nError)
	goto ErrorReturn;
  
    return mp;    

  ErrorReturn:    
    /*** anything to free? */
    /* ... */
    return NULL;
}



/*
** ExtendedAero_Calc ()
** 
** Return:
** 0	Ok
** <0	Not Ok, Error
**
*/

static int
ExtendedAero_Calc (
    void	*MP,		/* model handel/parameters		      */
    tAeroIF	*IF,		/* Interface structure			      */
    double	dt)		/* time step				      */
{                 
    struct tExtendedAero *mp = (struct tExtendedAero*)MP;
    double var;     
                  
    /* Calculation of the ride height:
     * Translation of the wheelcenter is in the Fr1 frame. This frame is 
     * connected to the vehicle body. If the wheelcenter moves up the body
     * moves down: tz_wheelcenter = -tz_body.(=Car.Susp[<pos>].t[2])
     * In addition the dynamical wheelradius and an offset value is
     * considered */
    rhf_old =  mp->rh_f;
    rhr_old =  mp->rh_r;
    rh_old  =  mp->rh  ;
    
    mp->rh_f = (float)(0.5 * (Car.Tire[0].WRadius - Car.Susp[0].t[2] 
 		+ Car.Tire[1].WRadius - Car.Susp[1].t[2])
		- mp->rh_offset_f);
    mp->rh_r = (float)(0.5 * (Car.Tire[2].WRadius - Car.Susp[2].t[2]
	    	+ Car.Tire[3].WRadius - Car.Susp[3].t[2])
	    	- mp->rh_offset_r);
    mp->rh = 0.5 * (mp->rh_f + mp->rh_r);
    
    /* Calculation of ride height velocity */
    mp->v_rhf = (mp->rh_f - rhf_old)/dt;
    mp->v_rhr = (mp->rh_r - rhr_old)/dt;
    mp->v_rh  = (mp->rh   - rh_old)/dt;
    
    /*************************************/
    /* Calculation of lift coefficients  */
    /* look up coefficients depending on the ride height */	
    mp->c_zf1 = LMEval(mp->Mapc_zf, mp->rh_f);
    mp->c_zr1 = LMEval(mp->Mapc_zr, mp->rh_r);
   
   /* look up coefficients depending on the ride height on both axles */
    mp->c_x2  = LM2DEval(mp->Map2Dc_x, mp->rh_f, mp->rh_r);
    mp->c_zf2 = LM2DEval(mp->Map2Dc_zf, mp->rh_f, mp->rh_r);
    mp->c_zr2 = LM2DEval(mp->Map2Dc_zr, mp->rh_f, mp->rh_r);
   
   /* Calculate sum of all drag coefficients  */
    mp->c_x_tot  = mp->c_x0 + mp->c_x2;
	   
    /* Calculate sum of all lift coefficients */
    mp->c_zf_tot = mp->c_zf0 + mp->c_zf1 + mp->c_zf2;
    mp->c_zr_tot = mp->c_zr0 + mp->c_zr1 + mp->c_zr2;
	 

    /* Calculation of drag and lift forces in x and z direction */ 
    /* var = 1/2 * rho * A * v */
    /*var = 0.5 * Ambient.AirDensity * mp->Ax*/
	var = 0.5 * 1.2 * mp->Ax
	    *IF->ApproachVel_1[0] * IF->ApproachVel_1[0];
    
    mp->Frc_x  = - var * mp->c_x_tot;
    /* Front axle */
    mp->Frc_zf =   var * mp->c_zf_tot;
    /* Rear axle */
    mp->Frc_zr =   var * mp->c_zr_tot;
    
#if 0
    if (SimCore.CycleNo % 5000 == 0){
	/*Log(" %f,  %f, %f \n ",mp->Drag.k1, mp->LiftF.k1,mp->LiftR.k1 );*/
	Log(" vrh:  %f,  %f, %f \n ",mp->v_rhf, mp->v_rhr ,mp->v_rh );
	Log(" dt: %f \n", dt);
    }
#endif

    /* Calculation and Assignment of CarMaker interface quantities */
    /* There is only one point of attack in CarMaker. The forces 
     * and torques on this point have to be calculated */

    /* point (x,y,z) of attack */
    IF->PoA_1[0] = mp->PoA_CM[0]; 
    IF->PoA_1[1] = mp->PoA_CM[1];
    IF->PoA_1[2] = mp->PoA_CM[2];
    
    /* forces to vehicle in x,y and z direction on the point of 
     * attack in CarMaker */
    IF->Frc_1[0] = mp->Frc_x;
    IF->Frc_1[1] = 0;
    IF->Frc_1[2] = mp->Frc_zf + mp->Frc_zr;
    
    /* torque to vehicle around x,y and z axle */
    /* */ 
    IF->Trq_1[0] = 0;
    IF->Trq_1[1] = -(mp->PoA_f[0] - mp->PoA_CM[0]) * mp->Frc_zf 
	    - (mp->PoA_r[0] - mp->PoA_CM[0]) * mp->Frc_zr;
    IF->Trq_1[2] = 0;

    return 0;
}



/*
** ExtendedAero_Delete ()
**
** free all allocated data
**
*/

static void
ExtendedAero_Delete (void *MP)
{
    struct tExtendedAero *mp = (struct tExtendedAero*)MP;

    free (mp);
}



/*
** Aero_Register_ExtendedAero ()
*/
int 
Aero_Register_ExtendedAero (void)
{
    tModelClassDescr m;

    Log("Register Model: %s", ThisModelKind);
    
    memset (&m, 0, sizeof(m));
    m.Aero.VersionId =			ThisVersionId;
    m.Aero.CompatVersionId =		ThisVersionId;
    m.Aero.New =			ExtendedAero_New;
    m.Aero.Calc =			ExtendedAero_Calc;
    m.Aero.Delete =			ExtendedAero_Delete;
    m.Aero.DeclQuants =		ExtendedAero_DeclQuants;
    
    return Model_Register(ModelClass_Aero, "ExtendedAero", &m);
}
